use egg::{AstSize, Extractor, Id, Pattern, PatternAst, RecExpr, Runner, Searcher};
use solveq::factorization::{FactorizationCost, FactorizationCostFn};
use solveq::language::{ConstantFold, EquationLanguage, Rational, RULES, EGraph};
use solveq::normal_form::extract_normal_form;
use solveq::parse::parse_equation;
use solveq::output::print_term;

static TEST_EQUATIONS: &[&str] = &[
	"(x + 50) * 10 - 150 = 100",
	"(x - 2) * (x + 2) = 0",
	"x ^ 2 = 4",
	"x ^ 2 - 2 = 0",
	"x ^ 2 = 2 * x + 15",
	"(x ^ 2 - 2 * x - 15) * (x + 5) = 0",
	"x ^ 3 + 3 * x ^ 2 - 25 * x - 75 = 0",
];


fn main() {
	//	let expr: RecExpr<EquationLanguage> = "(+ (* x (+ x -2)) -15)".parse().unwrap();
//	let expr: RecExpr<EquationLanguage> = "(* (+ (* x (+ x -2)) -15) (+ x 5))".parse().unwrap();
//	println!("{:?}", get_expression_cost(&expr));

	for eq in TEST_EQUATIONS {
		println!("Equation: {}", *eq);

		let mut start = parse_equation(*eq).unwrap();
		let root_id = Id::from(start.as_ref().len()-1);
		let EquationLanguage::Equals([left, right]) = start[root_id]
		else { panic!("Not an equation without an equals sign!"); };
		start[root_id] = EquationLanguage::Sub([left, right]);

		println!("Parsed: {}", &start);

		let mut runner = Runner::default()
			.with_explanations_enabled()
			.with_expr(&start)
			.run(&*RULES);

		let extractor = Extractor::new(&runner.egraph, FactorizationCostFn);

		let (best_cost, best_expr) = extractor.find_best(runner.roots[0]);

//		println!("Best expresssion: {} {:?}", best_expr, best_cost);
//		println!("{}", runner.explain_equivalence(&start, &best_expr).get_flat_string());

		let Some(factorization) = extract_normal_form(&runner.egraph, runner.roots[0]) else {
			panic!("Couldn't factorize polynomial!");
		};
		println!("Factorized normal form: {}", factorization);

		/*
		get_expression_cost("(* x (+ x -2))", &runner.egraph);
		get_expression_cost("-15", &runner.egraph);
		get_expression_cost("(+ (* x (+ x -2)) -15)", &runner.egraph);
		*/

//		let factorization = extract_factorization(&best_expr);


		let mut solutions: Vec<String> = Vec::new();
		for poly in &factorization.polynomials {
			if poly.len() == 2 { // linear factor
				let Rational { num, denom } = &poly[0];
				solutions.push(format!("x = {}", Rational { num: -*num, denom: *denom }));
			} else if poly.len() == 3 { // quadratic factor
				let Rational { num: num0, denom: denom0 } = &poly[0];
				let Rational { num: num1, denom: denom1 } = &poly[1];

				let sol1 = format!("- ({num1})/(2 * ({denom1})) + ((({num1})/(2 * ({denom1}))) ^ 2 - ({num0}) / ({denom0})) ^ (1/2)");
				let sol2 = format!("- ({num1})/(2 * ({denom1})) - ((({num1})/(2 * ({denom1}))) ^ 2 - ({num0}) / ({denom0})) ^ (1/2)");

				let expr = parse_equation(&sol1).unwrap();
				let runner = Runner::default()
					.with_expr(&expr)
					.run(&*RULES);
				let extractor = Extractor::new(&runner.egraph, AstSize);
				let (_, simplified_expr) = extractor.find_best(runner.roots[0]);
				solutions.push(format!("x = {}", print_term(&simplified_expr)));

				let expr = parse_equation(&sol2).unwrap();
				let runner = Runner::default()
					.with_expr(&expr)
					.run(&*RULES);
				let extractor = Extractor::new(&runner.egraph, AstSize);
				let (_, simplified_expr) = extractor.find_best(runner.roots[0]);
				solutions.push(format!("x = {}", print_term(&simplified_expr)));
			}
		}

		println!("Solutions: {{ {} }}", solutions.join(", "));
		println!("");
	}
}

fn get_expression_cost(expr: &str, egraph: &EGraph) {
//    let mut egraph = EGraph::new(ConstantFold::default());
	//    let id = egraph.add_expr(expr);
	let pattern: Pattern<EquationLanguage> = expr.parse().unwrap();
	let matches = pattern.search(egraph);
	for m in matches {
		let extractor = Extractor::new(&egraph, FactorizationCostFn);
		let (cost, _) = extractor.find_best(m.eclass);

		println!("expr: {}, id: {}, cost: {:?}", expr, m.eclass, cost);
	}
//    cost
}
